/*
 * Copyright (C) 2009 Emweb bvba, Leuven, Belgium.
 *
 * See the LICENSE file for terms of use.
 */
package eu.webtoolkit.jwt;

import java.util.EnumSet;
import java.util.List;

/**
 * A label for a form field.
 * <p>
 * 
 * The label may contain an image and/or text. It acts like a proxy for giving
 * focus to a {@link WFormWidget}. When both an image and text are specified,
 * the image is put to the left of the text.
 * <p>
 * Usage example:
 * <p>
 * <blockquote>
 * 
 * <pre>
 * WContainerWidget w = new WContainerWidget();
 * WLabel label = new WLabel(&quot;Favourite Actress: &quot;, w);
 * WLineEdit edit = new WLineEdit(&quot;Renee Zellweger&quot;, w);
 * label.setBuddy(edit);
 * </pre>
 * 
 * </blockquote>
 * <p>
 * The widget corresponds to the HTML <code>&lt;label&gt;</code> tag. When no
 * buddy is set, it is rendered using an HTML <code>&lt;span&gt;</code> or
 * <code>&lt;div&gt;</code> to avoid click event handling misbehavior on
 * Microsoft Internet Explorer.
 * <p>
 * WLabel is an {@link WWidget#setInline(boolean inlined) inline} widget.
 * <p>
 * <h3>CSS</h3>
 * <p>
 * This widget does not provide styling, and can be styled using inline or
 * external CSS as appropriate. A label&apos;s text may be styled via a nested
 * <code>&lt;span&gt;</code> element, and it&apos;s image may be styled via a
 * nested <code>&lt;img&gt;</code> element.
 */
public class WLabel extends WInteractWidget {
	/**
	 * Creates a label with empty text and optional parent.
	 */
	public WLabel(WContainerWidget parent) {
		super(parent);
		this.buddy_ = null;
		this.text_ = null;
		this.image_ = null;
		this.buddyChanged_ = false;
		this.newImage_ = false;
		this.newText_ = false;
	}

	/**
	 * Creates a label with empty text and optional parent.
	 * <p>
	 * Calls {@link #WLabel(WContainerWidget parent)
	 * this((WContainerWidget)null)}
	 */
	public WLabel() {
		this((WContainerWidget) null);
	}

	/**
	 * Creates a label with a given text.
	 */
	public WLabel(CharSequence text, WContainerWidget parent) {
		super(parent);
		this.buddy_ = null;
		this.image_ = null;
		this.buddyChanged_ = false;
		this.newImage_ = false;
		this.newText_ = false;
		this.text_ = new WText(text);
		this.text_.setWordWrap(false);
		this.text_.setParentWidget(this);
	}

	/**
	 * Creates a label with a given text.
	 * <p>
	 * Calls {@link #WLabel(CharSequence text, WContainerWidget parent)
	 * this(text, (WContainerWidget)null)}
	 */
	public WLabel(CharSequence text) {
		this(text, (WContainerWidget) null);
	}

	/**
	 * Creates a label with an image.
	 */
	public WLabel(WImage image, WContainerWidget parent) {
		super(parent);
		this.buddy_ = null;
		this.text_ = null;
		this.buddyChanged_ = false;
		this.newImage_ = false;
		this.newText_ = false;
		this.image_ = image;
		this.image_.setParentWidget(this);
	}

	/**
	 * Creates a label with an image.
	 * <p>
	 * Calls {@link #WLabel(WImage image, WContainerWidget parent) this(image,
	 * (WContainerWidget)null)}
	 */
	public WLabel(WImage image) {
		this(image, (WContainerWidget) null);
	}

	public void remove() {
		this.setBuddy((WFormWidget) null);
		super.remove();
	}

	/**
	 * Returns the buddy of this label.
	 * <p>
	 * 
	 * @see WLabel#setBuddy(WFormWidget buddy)
	 */
	public WFormWidget getBuddy() {
		return this.buddy_;
	}

	/**
	 * Sets the buddy of this label.
	 * <p>
	 * Sets the buddy FormWidget for which this label acts as a proxy.
	 * <p>
	 * 
	 * @see WFormWidget#getLabel()
	 * @see WLabel#getBuddy()
	 */
	public void setBuddy(WFormWidget buddy) {
		if (this.buddy_ != null) {
			this.buddy_.setLabel((WLabel) null);
		}
		this.buddy_ = buddy;
		if (this.buddy_ != null) {
			this.buddy_.setLabel(this);
		}
		this.buddyChanged_ = true;
		this.repaint(EnumSet.of(RepaintFlag.RepaintPropertyAttribute));
	}

	/**
	 * Sets the label text.
	 */
	public void setText(CharSequence text) {
		if (this.getText().equals(text)) {
			return;
		}
		if (!(this.text_ != null)) {
			this.text_ = new WText();
			this.text_.setWordWrap(false);
			this.text_.setParentWidget(this);
			this.newText_ = true;
			this.repaint(EnumSet.of(RepaintFlag.RepaintInnerHtml));
		}
		this.text_.setText(text);
	}

	/**
	 * Returns the label text.
	 */
	public WString getText() {
		if (this.text_ != null) {
			return this.text_.getText();
		} else {
			return empty;
		}
	}

	/**
	 * Sets the image.
	 */
	public void setImage(WImage image, Side side) {
		if (this.image_ != null)
			this.image_.remove();
		this.image_ = image;
		if (this.image_ != null) {
			this.image_.setParentWidget(this);
			this.imageSide_ = side;
		}
		this.newImage_ = true;
		this.repaint(EnumSet.of(RepaintFlag.RepaintInnerHtml));
	}

	/**
	 * Sets the image.
	 * <p>
	 * Calls {@link #setImage(WImage image, Side side) setImage(image,
	 * Side.Left)}
	 */
	public final void setImage(WImage image) {
		setImage(image, Side.Left);
	}

	/**
	 * Returns the image.
	 */
	public WImage getImage() {
		return this.image_;
	}

	/**
	 * Configures word wrapping.
	 * <p>
	 * When <code>wordWrap</code> is <code>true</code>, the widget may break
	 * lines, creating a multi-line text. When <code>wordWrap</code> is
	 * <code>false</code>, the text will displayed on a single line, unless the
	 * text contains end-of-lines (for {@link TextFormat#PlainText}) or &lt;br
	 * /&gt; tags or other block-level tags (for {@link TextFormat#XHTMLText}).
	 * <p>
	 * The default value is <code>false</code>.
	 * <p>
	 * 
	 * @see WLabel#hasWordWrap()
	 */
	public void setWordWrap(boolean wordWrap) {
		if (!(this.text_ != null)) {
			this.text_ = new WText();
			this.text_.setParentWidget(this);
			this.newText_ = true;
			this.repaint(EnumSet.of(RepaintFlag.RepaintInnerHtml));
		}
		this.text_.setWordWrap(wordWrap);
	}

	/**
	 * Returns whether word wrapping is on.
	 * <p>
	 * 
	 * @see WLabel#setWordWrap(boolean wordWrap)
	 */
	public boolean hasWordWrap() {
		return this.text_ != null ? this.text_.isWordWrap() : false;
	}

	private WFormWidget buddy_;
	private WText text_;
	private WImage image_;
	private Side imageSide_;
	private boolean buddyChanged_;
	private boolean newImage_;
	private boolean newText_;

	void updateDom(DomElement element, boolean all) {
		WApplication app = WApplication.getInstance();
		if (this.image_ != null && this.text_ != null) {
			if (this.imageSide_ == Side.Left) {
				this.updateImage(element, all, app, 0);
				this.updateText(element, all, app, 1);
			} else {
				this.updateText(element, all, app, 0);
				this.updateImage(element, all, app, 1);
			}
		} else {
			this.updateText(element, all, app, 0);
			this.updateImage(element, all, app, 0);
		}
		if (this.buddyChanged_ || all) {
			if (this.buddy_ != null) {
				element.setAttribute("for", this.buddy_.getFormName());
			}
			this.buddyChanged_ = false;
		}
		super.updateDom(element, all);
	}

	DomElementType getDomElementType() {
		if (this.buddy_ != null) {
			return DomElementType.DomElement_LABEL;
		} else {
			return this.isInline() ? DomElementType.DomElement_SPAN
					: DomElementType.DomElement_DIV;
		}
	}

	void getDomChanges(List<DomElement> result, WApplication app) {
		super.getDomChanges(result, app);
		if (this.text_ != null) {
			((WWebWidget) this.text_).getDomChanges(result, app);
		}
		if (this.image_ != null) {
			((WWebWidget) this.image_).getDomChanges(result, app);
		}
	}

	void propagateRenderOk(boolean deep) {
		this.newImage_ = false;
		this.newText_ = false;
		this.buddyChanged_ = false;
		super.propagateRenderOk(deep);
	}

	protected void propagateSetEnabled(boolean enabled) {
		if (this.text_ != null) {
			if (enabled) {
				this.text_.removeStyleClass("Wt-disabled");
			} else {
				this.text_.addStyleClass("Wt-disabled");
			}
		}
		super.propagateSetEnabled(enabled);
	}

	protected void updateImage(DomElement element, boolean all,
			WApplication app, int pos) {
		if (this.newImage_ || all) {
			if (this.image_ != null) {
				element.insertChildAt(((WWebWidget) this.image_)
						.createDomElement(app), pos);
			}
			this.newImage_ = false;
		}
	}

	protected void updateText(DomElement element, boolean all,
			WApplication app, int pos) {
		if (this.newText_ || all) {
			if (this.text_ != null) {
				element.insertChildAt(((WWebWidget) this.text_)
						.createDomElement(app), pos);
			}
			this.newText_ = false;
		}
	}

	static WString empty = new WString("");
}
